use crate::{felinet::handler_client::HandlerClient, util::validate_freq};

use anyhow::{anyhow, bail, Context};
use clap::{arg, Parser};
use config::Config;
use felinet::PingRequest;
use gate::{beacon_forever, felinet_receive_forever, felinet_send_forever, packet_handler};
use ham_cats::{
    buffer::Buffer,
    packet::Packet,
    whisker::{Arbitrary, Identification},
};
use radio::{RadioManager, MAX_PACKET_LEN};
use std::{str::FromStr, time::Duration};
use tokio::sync::{broadcast, mpsc};
use tonic::{transport::Endpoint, Request};
use uuid::Uuid;

mod config;
mod gate;
mod radio;
mod util;

pub mod felinet {
    tonic::include_proto!("felinet");
}

#[derive(Parser)]
struct Cli {
    /// Repeatedly send very long packets on the specified frequency for debugging purposes.
    /// Can be used when measuring power output, for example.
    /// Valid frequencies are 420MHz - 450MHz
    #[arg(long, short, name = "freq")]
    xmit_test: Option<f64>,
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    let config = Config::load()?;
    let cli = Cli::parse();

    let (packet_tx, packet_receive) = mpsc::channel(16);
    let (packet_send, packet_rx) = mpsc::channel(16);
    let radio = RadioManager::new(packet_tx, packet_rx, config.frequency)
        .expect("Could not initialize radio");

    match cli.xmit_test {
        Some(freq) => xmit_test(&config, freq, radio, packet_send).await?,
        None => gate_forever(&config, radio, packet_send, packet_receive).await?,
    }

    Ok(())
}

async fn xmit_test(
    config: &Config,
    freq: f64,
    mut radio: RadioManager,
    packet_send: mpsc::Sender<Vec<u8>>,
) -> anyhow::Result<()> {
    let freq_hz = (freq * 1_000_000.0) as u32;

    if !validate_freq(freq_hz) {
        bail!("Frequency {:.3} MHz out of range (420MHz - 450MHz)", freq);
    }

    radio.set_frequency(freq_hz)?;

    let mut buf = [0; MAX_PACKET_LEN];
    let mut pkt = Packet::new(&mut buf);
    pkt.add_identification(
        Identification::new(&config.callsign, config.ssid, config.icon)
            .context("Invalid identification")?,
    )
    .map_err(|e| anyhow!("Could not add identification to packet: {e}"))?;
    pkt.add_comment("Debugging packet")
        .map_err(|e| anyhow!("Could not add comment to packet: {e}"))?;

    for _ in 0..((MAX_PACKET_LEN - 100) / 600) {
        pkt.add_arbitrary(Arbitrary::new(&[0xA5; 255]).unwrap())
            .map_err(|e| anyhow!("Could not add data to packet: {e}"))?;
    }

    let mut buf2 = [0; MAX_PACKET_LEN];
    let mut data = Buffer::new_empty(&mut buf2);
    pkt.fully_encode(&mut data)
        .map_err(|e| anyhow!("Could not encode packet: {e}"))?;

    tokio::task::spawn(async move {
        loop {
            if let Err(e) = radio.process_forever().await {
                eprintln!("Radio error: {e}")
            }
        }
    });

    println!("Transmitting test data on {freq:.3} MHz... Press Ctrl+C to stop");

    loop {
        packet_send.send(data.to_vec()).await?;
    }
}

async fn gate_forever(
    config: &Config,
    mut radio: RadioManager,
    packet_send: mpsc::Sender<Vec<u8>>,
    packet_receive: mpsc::Receiver<(Vec<u8>, f64)>,
) -> anyhow::Result<()> {
    let uuid = Uuid::new_v4();

    let temperature_mutex = radio.temperature_mutex();

    let tx = broadcast::Sender::new(16);

    beacon_forever(
        config,
        packet_send.clone(),
        tx.clone(),
        uuid,
        temperature_mutex,
    )?;

    let client = if let Some(felinet_config) = &config.felinet {
        let endpoint = Endpoint::from_str(&felinet_config.address)?
            .keep_alive_while_idle(true)
            .keep_alive_timeout(Duration::from_secs(5))
            .http2_keep_alive_interval(Duration::from_secs(5))
            .tcp_keepalive(Some(Duration::from_secs(5)));

        Some(HandlerClient::connect(endpoint).await?)
    } else {
        eprintln!(
            "No FELINET section specified in configuration. Received packets will not be uploaded!"
        );
        None
    };

    let callsign = config.callsign.clone();

    {
        let tx = tx.clone();
        let packet_send = packet_send.clone();
        let ssid = config.ssid;
        let callsign = config.callsign.clone();

        tokio::task::spawn(async move {
            if let Err(e) = packet_handler(
                callsign.clone(),
                ssid,
                tx,
                packet_send,
                packet_receive,
                uuid,
            )
            .await
            {
                eprintln!("Packet handler error: {}. Aborting", e)
            }
        });
    }

    if let Some(client) = client {
        {
            let client = client.clone();
            tokio::task::spawn(async {
                felinet_send_forever(client, tx).await;
            });
        }

        {
            let mut client = client.clone();
            tokio::task::spawn(async move {
                loop {
                    client
                        .ping(Request::new(PingRequest { uuid: uuid.into() }))
                        .await
                        .ok();

                    tokio::time::sleep(Duration::from_secs(600)).await;
                }
            });
        }

        let ssid = config.ssid;
        tokio::task::spawn(async move {
            felinet_receive_forever(client, packet_send, callsign, ssid, uuid).await;
        });
    }

    loop {
        if let Err(e) = radio.process_forever().await {
            eprintln!("Radio error: {e}")
        }
    }
}
